# -*- coding: utf-8 -*-

#########################################################
# Some Logging and Plotting Candies to Monitor Training
#########################################################

import time
import math

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np


def asMinutes(s):
    m = math.floor(s / 60)
    s -= m * 60
    return '%dm %ds' % (m, s)

def timeSince(since, percent):
    now = time.time()
    s = now - since
    es = s / (percent)
    rs = es - s
    return '%s (- %s)' % (asMinutes(s), asMinutes(rs))


def showPlot(points):
    plt.figure()
    fig, ax = plt.subplots()
    # this locator puts ticks at regular intervals
    loc = ticker.MultipleLocator(base=0.2)
    ax.yaxis.set_major_locator(loc)
    plt.plot(points)


#########################################################
# Training per epoch,
# Iterates across data points per epoch.
#########################################################
def train_one_epoch(input_variable, target_variable, encoder, decoder,
                    encoder_optimizer, decoder_optimizer, criterion):
    """
    Function to put the variables, decoder and optimizers to train per epoch.
    """
    encoder_hidden = encoder.initialize_hidden_states()

    # (Re-)Initialize the optimizers, clear all gradients.
    encoder_optimizer.zero_grad()
    decoder_optimizer.zero_grad()

    # Initialize the length of the PyTorch variables.
    input_length = input_variable.size()[0]
    target_length = target_variable.size()[0]
    encoder_outputs = Variable(torch.zeros(MAX_LENGTH, encoder.hidden_size))
    encoder_outputs = encoder_outputs.cuda() if use_cuda else encoder_outputs

    loss = 0

    # Iterating through each word in the input.
    for ei in range(input_length):
        # We move forward through each state.
        encoder_output, encoder_hidden = encoder(input_variable[ei], encoder_hidden)
        # And we save the encoder outputs.
        encoder_outputs[ei] = encoder_output[0][0]

    # Initialize the variable input with the index of the START.
    decoder_input = Variable(torch.LongTensor([[START_IDX]]))
    decoder_input = decoder_input.cuda() if use_cuda else decoder_input

    # As the first state of the decoder, we take the last step of the encoder.
    decoder_hidden = encoder_hidden

    # Without teacher forcing: use its own predictions as the next input
    for di in range(target_length):
        decoder_output, decoder_hidden = decoder(
            decoder_input, decoder_hidden)
        topv, topi = decoder_output.data.topk(1)
        ni = topi[0][0]

        decoder_input = Variable(torch.LongTensor([[ni]]))
        decoder_input = decoder_input.cuda() if use_cuda else decoder_input

        loss += criterion(decoder_output, target_variable[di])
        if ni == END_IDX:
            break

    loss.backward()

    encoder_optimizer.step()
    decoder_optimizer.step()

    return loss.data[0] / target_length



#########################################################
# Top-level function to start the training,
# iterates across epochs.
#########################################################
def train(encoder, decoder, n_iters,  learning_rate=0.01, print_every=1000, plot_every=100,):
    start = time.time()
    plot_losses = []
    print_loss_total = 0  # Reset every print_every
    plot_loss_total = 0  # Reset every plot_every

    encoder_optimizer = optim.SGD(encoder.parameters(), lr=learning_rate)
    decoder_optimizer = optim.SGD(decoder.parameters(), lr=learning_rate)

    training_pairs = [random.choice(sent_pairs) for i in range(n_iters)]

    criterion = nn.NLLLoss()

    for iter in range(1, n_iters + 1):
        training_pair = training_pairs[iter - 1]
        input_variable = training_pair[0]
        target_variable = training_pair[1]

        loss = train_one_epoch(input_variable, target_variable, encoder,
                     decoder, encoder_optimizer, decoder_optimizer, criterion)
        print_loss_total += loss
        plot_loss_total += loss

        if iter % print_every == 0:
            print_loss_avg = print_loss_total / print_every
            print_loss_total = 0
            print('%s (%d %d%%) %.4f' % (timeSince(start, iter / n_iters),
                                         iter, iter / n_iters * 100, print_loss_avg))

        if iter % plot_every == 0:
            plot_loss_avg = plot_loss_total / plot_every
            plot_losses.append(plot_loss_avg)
            plot_loss_total = 0

    showPlot(plot_losses)



# -*- coding: utf-8 -*-

#########################################################
# Some Logging and Plotting Candies to Monitor Training
#########################################################

import time
import math

import matplotlib.pyplot as plt
import matplotlib.ticker as ticker
import numpy as np


def asMinutes(s):
    m = math.floor(s / 60)
    s -= m * 60
    return '%dm %ds' % (m, s)

def timeSince(since, percent):
    now = time.time()
    s = now - since
    es = s / (percent)
    rs = es - s
    return '%s (- %s)' % (asMinutes(s), asMinutes(rs))


def showPlot(points):
    plt.figure()
    fig, ax = plt.subplots()
    # this locator puts ticks at regular intervals
    loc = ticker.MultipleLocator(base=0.2)
    ax.yaxis.set_major_locator(loc)
    plt.plot(points)



# -*- coding: utf-8 -*-

import numpy as np

from torch.utils.data.dataset import Dataset
from torch.utils.data import DataLoader

from komorebi.parallel import ParallelData
from komorebi.util import absolute_path, timing

class ParallelDataset(Dataset):
    def __init__(self, src_file, trg_file, src_max_len=None, trg_max_len=None,
                 **kwargs):
        """
        This is the object to be used by PyTorch.
        """
        # ParallelData object is an iterator packe with goodies =)
        self._data = ParallelData(src_file, trg_file, **kwargs)
        # Iterate through the dataset to get the source and data.
        _source_texts, _target_texts = zip(*self._data)
        self._data.reset() # Resets the iterator to the start.
        # Pre-computes the no. of data points.
        self._len = len(_source_texts)
        # Precompute the real leangth of the source and target text.
        self.source_lens = list(map(len, _source_texts))
        self.target_lens = list(map(len, _target_texts))
        # Set the max lengths if not specified by user.
        self.src_max_len = src_max_len if src_max_len else max(self.source_lens)
        self.trg_max_len = trg_max_len if trg_max_len else max(self.target_lens)
        # Converts the indices to torch tensors.
        pad_idx = self._data.src_data.PAD_IDX
        self.source_texts = np.array([self.pad_sequence(_s, self.src_max_len, pad_idx)
                                      for _s in _source_texts])
        self.target_texts = np.array([self.pad_sequence(_t, self.trg_max_len, pad_idx)
                                      for _t in _target_texts])

    def pad_sequence(self, sequence, max_len, pad_idx):
        padded_sequence = np.zeros(max_len, dtype=np.int64)
        padded_sequence[:len(sequence)] = sequence[:max_len]
        if pad_idx != 0:
            padded_sequence[len(x):] = pad_idx
        return padded_sequence

    def __len__(self):
        return self._len

    def __getitem__(self, index):
        return {'x': self.source_texts[index],
                'x_len': np.clip(self.source_lens[index], 1, self.src_max_len),
                'y': self.target_texts[index]}

    def batches(self, batch_size, shuffle=True):
        dataloader = DataLoader(dataset=self, batch_size=batch_size, shuffle=shuffle)
        for data_dict in dataloader:
            # Sort indices of data in batch by lengths.
            sorted_indices = np.array(data_dict['x_len']).argsort()[::-1].tolist()
            data_batch = {name:_tensor[sorted_indices]
                          for name, _tensor in data_dict.items()}
            yield data_batch

    @timing
    def save(self, saveto, save_tensors=True):
        pass

    @timing
    def load():
        pass



# -*- coding: utf-8 -*-

import heapq
import itertools
import json
import os
import pickle
import random
import sys

from collections import Iterator
from itertools import chain
from pathlib import Path
from operator import itemgetter

from komorebi.text import TextData
from komorebi.util import per_chunk, timing
from komorebi.util import DataError

class ParallelData(Iterator):
    def __init__(self,
                 src_filename=None,
                 trg_filename=None,
                 src_vocab_size=10**5,
                 trg_vocab_size=None,
                 chunk_size=10**5,
                 delimiter=None,
                 size_mb=4024,
                 pad_symbol='<pad>',
                 start_symbol='<s>',
                 end_symbol='</s>',
                 unknown_symbol='<unk>',
                 default_pad_start=False,
                 default_pad_end=True,
                 filter_on='tf',
                 prune_at=10**10,
                 encoding='utf8',
                 **kwargs):
        """
        This is the object to store parallel text and read them into vocabulary
        indices. The object is an iterable that yields tuples of the vocabulary
        indices, one from the source sentence, another from the target.

        :param src_filename: Textfile that contains source sentences.
        :type src_filename: str
        :param trg_filename: Textfile that contains target sentences.
        :type trg_filename: str
        :param src_vocab_size: Max no. of words to keep in the source vocab.
        :type src_vocab_size: int
        :param trg_vocab_size: Max no. of words to keep in the target vocab.
        :type trg_vocab_size: int
        :param chunk_size: Use to limit no. of sentences to load at a time when populating the vocabulary.
        :type chunk_size: int
        :param delimiter: Delimiter to split on when "tokenizing"
        :type delimiter: str
        :param size_mb: Memory footprint of the bounter object use to count the vocab.
        :type size_mb: int
        :param start_symbol: Start symbol use for padding.
        :type start_symbol: str
        :param end_symbol: End symbol use for padding.
        :type end_symbol: str
        :param unknown_symbol: Unknown symbol for OOV words.
        :type unknown_symbol: str
        :param filter_on: Option to filter on term-freq ('tf') or doc-freq ('df')
        :type filter_on: str
        :param prune_at: *prune_at* parameter used by gensim.Dictionary
        :type prune_at: int
        """

        if 'loadfrom' not in kwargs: # Creating.
            self.default_pad_start = default_pad_start
            self.default_pad_end = default_pad_end
            print('Creating source TextData...', end='\n', file=sys.stderr)
            self.src_data = TextData(src_filename, src_vocab_size, **kwargs)
            print('Creating target TextData...', end='\n', file=sys.stderr)
            self.trg_data = TextData(trg_filename, trg_vocab_size, **kwargs)
            self.iterable = self._iterate()

        else: # Loading.
            self.load(kwargs['loadfrom'],
                      src_filename, trg_filename,
                      kwargs.get('load_counter', False))
            self.iterable = self._iterate()

    @timing
    def load(self, loadfrom, src_filename=None, trg_filename=None,
             load_counter=False):
        """
        The load function.
        :param loadfrom: The path to load the directory for the ParallelData.
        :type loadfrom: str
        :param load_counter: Whether to load the src and trg bounter objects.
        :type load_counter: bool
        """
        config_file = loadfrom + '/ParallelData.json'
        if not Path(config_file).exists():
            raise DataError('{} config file not found!!'.format(config_file))
        else:
            print('Loading ParallelData from {}'.format(config_file),
                  end=' ', file=sys.stderr)
            with open(config_file) as fin:
                self.__dict__ = json.load(fin)

            if ('src_data' not in self.__dict__ or
                'trg_data' not in self.__dict__):
                raise DataError('source/target TextData not found!!')

            json.load(open(config_file))['src_data']
            json.load(open(config_file))['trg_data']

            # Actual loading of TextData objects.
            textdatas = {'src_data': src_filename, 'trg_data': trg_filename}
            for textdata, filename in textdatas.items():
                try:
                    self.__dict__[textdata] = TextData(loadfrom=self.__dict__[textdata],
                                                       filename=filename,
                                                       load_counter=load_counter)
                except DataError as e:
                    raise DataError("You need to set the src_filename and trg_filename when loading TextData, e.g.\n"
                                    "\tParallelData(loadfrom='path/to/textdata', \n"
                                    "\t\t\tsrc_filename='srcfile.txt', \n"
                                    "\t\t\ttrg_filename='trgfile.txt')")

    @timing
    def save(self, saveto, save_counter=False, copy_data=True):
        """
        The save function.
        :param saveto: The path to save the directory for the ParallelData.
        :type saveto: str
        :param save_counter: Whether to save the src and trg bounter objects.
        :type save_counter: bool
        :para copy_data: Make a local copy of the data.
        :type copy_data: bool
        """
        print("Saving ParallelData to {saveto}".format(saveto=saveto), end='\n', file=sys.stderr)
        # Create the directory if it doesn't exist.
        if not Path(saveto).exists():
            os.makedirs(saveto)
            os.makedirs(saveto+'/src/')
            os.makedirs(saveto+'/trg/')

        self.src_data.save(saveto+'/src/', save_counter, copy_data)
        self.trg_data.save(saveto+'/trg/', save_counter, copy_data)

        # Initialize the config file.
        config_json = {'src_data': saveto+'/src/',
                       'trg_data': saveto+'/trg/'}

        # Dump the config file.
        with open(saveto+'/ParallelData.json', 'w') as fout:
            json.dump(config_json, fout, indent=2)
        print("Saving ParallelData to {saveto}".format(saveto=saveto), end=' ', file=sys.stderr)

    def reset(self):
        """
        Resets the iterator to the 0th item.
        """
        self.iterable = self._iterate()

    def _iterate(self):
        """
        The helper function to iterate through the source and target file
        and convert the lines into vocabulary indices.
        """
        for src_line, trg_line in zip(self.src_data.lines(), self.trg_data.lines()):
            src_sent = self.src_data.vectorize(src_line,
                                               self.default_pad_start,
                                               self.default_pad_end)
            trg_sent = self.trg_data.vectorize(trg_line,
                                               self.default_pad_start,
                                               self.default_pad_end)
            yield src_sent, trg_sent

    def __next__(self):
        return next(self.iterable)

    def shuffle(self):
        return iter(sorted(self, key=lambda k: random.random()))



# -*- coding: utf-8 -*-

import heapq
import itertools
import json
import os
import pickle
import random
import sys

from shutil import copyfile
from collections import Iterator
from itertools import chain
from pathlib import Path
from operator import itemgetter

from tqdm import tqdm

from gensim.corpora.dictionary import Dictionary
from bounter import bounter

from komorebi.util import absolute_path, per_chunk, timing
from komorebi.util import DataError

class TextData(Iterator):
    def __init__(self,
                 filename=None,
                 vocab_size=None,
                 max_len=None,
                 chunk_size=10**5,
                 delimiter=None,
                 size_mb=4024,
                 pad_symbol='<pad>',
                 start_symbol='<s>',
                 end_symbol='</s>',
                 unknown_symbol='<unk>',
                 default_pad_start=False,
                 default_pad_end=True,
                 filter_on=None,
                 prune_at=10**10,
                 encoding='utf8',
                 **kwargs):
        """
        This is the object to store text and read them into vocabulary
        indices. The object is an iterable that yields vocabulary indices of the
        tokens in the sentences.
        :param filename: Textfile that contains source sentences.
        :type filename: str
        :param vocab_size: Max no. of words to keep in the source vocab.
        :type vocab_size: int
        :param chunk_size: Use to limit no. of sentences to load at a time when populating the vocabulary.
        :type chunk_size: int
        :param delimiter: Delimiter to split on when "tokenizing"
        :type delimiter: str
        :param size_mb: Memory footprint of the bounter object use to count the vocab.
        :type size_mb: int
        :param start_symbol: Start symbol use for padding.
        :type start_symbol: str
        :param end_symbol: End symbol use for padding.
        :type end_symbol: str
        :param unknown_symbol: Unknown symbol for OOV words.
        :type unknown_symbol: str
        :param default_pad_start: By default, pad the <s> to sentence when vectorizing.
        :type default_pad_start: bool
        :param default_pad_end: By default, pad the </s> to sentence when vectorizing.
        :type default_pad_end: bool
        :param filter_on: Option to filter on term-freq ('tf') or doc-freq ('df')
        :type filter_on: str
        :param prune_at: *prune_at* parameter used by gensim.Dictionary
        :type prune_at: int
        """
        if 'loadfrom' not in kwargs: # Creating.

            self.filename = absolute_path(filename)

            # Check that inputs are not None.
            assert Path(self.filename).exists(), "File {filename} does not exist".format(filename=filename)

            # Initialize encoding.
            self.encoding = encoding

            # Initialize the pad, start, end and unknown symbols.
            self.PAD, self.PAD_IDX = pad_symbol, 0
            self.START, self.START_IDX = start_symbol, 1
            self.END, self.END_IDX = end_symbol, 2
            self.UNK, self.UNK_IDX = unknown_symbol, 3
            self.default_pad_start = default_pad_start
            self.default_pad_end = default_pad_end

            # Save the user-specific delimiter
            self.delimiter = delimiter

            # Gensim related attribute to keep the pruning cap.
            self.prune_at = prune_at

            # Populate the source vocabulary.
            print('Creating Vocabulary...', end='\n', file=sys.stderr)
            self.vocab = Dictionary([[pad_symbol], [start_symbol], [end_symbol], [unknown_symbol]],
                                     prune_at=self.prune_at)
            self.counter = bounter(size_mb=size_mb)

            print('Building source vocab and counter...', end=' ', file=sys.stderr)
            self.populate_dictionary(self.filename, self.vocab,
                                     self.counter, chunk_size)
            # Use the user-specified source/target vocab size if set,
            # else use the full vocab_size.
            self.vocab_size = min(len(self.vocab), vocab_size) if vocab_size else len(self.vocab)

            # Keep the vocabulary to a max set by user.
            if filter_on and self.vocab_size < len(self.vocab):
                print('Filtering least frequent words in vocab.', end='\n', file=sys.stderr)
                if filter_on == 'tf':
                    self.filter_n_least_frequent(self.vocab,
                                                 self.counter,
                                                 self.vocab_size,
                                                 keep_tokens=['<pad>', '<s>', '</s>', '<unk>'])
                elif filter_on == 'df':
                    self.vocab.filter_extremes(no_below=1, no_above=self.prune_at,
                                               keep_n=self.vocab_size,
                                               keep_tokens=['<pad>', '<s>', '</s>', '<unk>'])

            self.iterable = self._iterate()

        else: # Loading.
            self.load(kwargs['loadfrom'], filename,
                      kwargs.get('load_counter', False))
            self.iterable = self._iterate()

    @timing
    def load(self, loadfrom=None, filename=None, load_counter=False):
        """
        The load function.

        :param filename: Path to the filename of the corpus to read,
                         this will overwrite filename in the TextData.json.
        :type filename: str

        :param loadfrom: The path to load the directory for the ParallelData.
        :type loadfrom: str

        :param load_counter: Whether to load the src and trg bounter objects.
        :type load_counter: bool
        """
        assert loadfrom is not None
        config_file = loadfrom + '/TextData.json'
        if not Path(config_file).exists():
            raise DataError('{} config file not found!!'.format(config_file))
        else:
            print('Loading TextData from {}'.format(config_file),
                  end=' ', file=sys.stderr)
            with open(config_file) as fin:
                self.__dict__ = json.load(fin)

            # If the data is saved with TextData.save(copy_data=True),
            # it will appear in self.__dict__ and
            # we set the filename from relative to absolute path
            # if data is copied when saved, i.e. `filename` in self.__dict__
            if 'filename' in self.__dict__:
                self.filename = os.path.join(loadfrom, self.filename)
            # If user specified filename when loading the TextData, e.g.
            #   TextData(filename='path/to/textfile', loadfrom='...'),
            # then we overwrite the filename.
            elif filename:
                self.filename = filename
            else:
                raise DataError("You need to set the filename when loading TextData, e.g.\n"
                                "\tTextData(loadfrom='path/to/textdata', filename='inputfile.txt')")
            # Check if the filename exists.
            if not os.path.isfile(self.filename):
                raise DataError("The text file at {} doesn't exist!!")

            try:
                with open(os.path.join(loadfrom, self.vocab), 'rb') as fin:
                    self.vocab = pickle.load(fin)
            except:
                raise DataError("{}/vocab.pkl isn't found".format(loadfrom))

            if load_counter:
                if ('counter' not in self.__dict__):
                    raise DataError('TextData counter not found!!')
                with open(os.path.join(loadfrom, self.counter), 'rb') as fin:
                    self.counter = pickle.load(fin)

    @timing
    def save(self, saveto, save_counter=False, copy_data=False):
        """
        The save function.
        :param saveto: The path to save the directory for the TextData.
        :type saveto: str
        :param save_counter: Whether to save the bounter objects.
        :type save_counter: bool
        :para copy_data: Make a local copy of the data.
        :type copy_data: bool
        """
        print("Saving TextData to {saveto}".format(saveto=saveto), end=' ', file=sys.stderr)
        # Create the directory if it doesn't exist.
        if not Path(saveto).exists():
            os.makedirs(saveto)

        # Save the vocab files.
        with open(saveto+'/vocab.pkl', 'wb') as fout:
            pickle.dump(self.vocab, fout)
        with open(saveto+'/vocab.tsv', 'w') as fout:
            for idx, word in self.vocab.items():
                print('\t'.join([str(idx), word]), end='\n', file=fout)

        # Initialize the config file.
        config_json = {'delimiter': self.delimiter, 'encoding': self.encoding,
                       'PAD': self.PAD, 'PAD_IDX': self.PAD_IDX,
                       'START': self.START, 'START_IDX': self.START_IDX,
                       'END': self.END, 'END_IDX': self.END_IDX,
                       'UNK': self.UNK, 'UNK_IDX': self.UNK_IDX,
                       'vocab_size': self.vocab_size,
                       'vocab': 'vocab.pkl',
                       'default_pad_start': self.default_pad_start,
                       'default_pad_end': self.default_pad_end}

        # Check whether we should save the counter.
        if save_counter:
            with open(saveto+'/counter.pkl', 'wb') as fout:
                pickle.dump(self.counter, fout)
            with open(saveto+'/counter.tsv', 'w') as fout:
                for word, count in self.counter.items():
                    print('\t'.join([str(word), str(count)]), end='\n', file=fout)

        config_json['counter'] = 'counter.pkl' if save_counter else None

        if copy_data:
            _, _filename = os.path.split(self.filename) # Filename without path.
            new_filename = os.path.join(saveto, _filename)
            print('\n\tCopying {} \n\tto {}'.format(self.filename, new_filename),
                  end='\n', file=sys.stderr)
            copyfile(absolute_path(self.filename), new_filename)
            config_json['filename'] = _filename

        # Dump the config file.
        with open(saveto+'/TextData.json', 'w') as fout:
            json.dump(config_json, fout, indent=2)

    def split_tokens(self, s):
        """
        A "tokenizer" that splits on space. If the delimiter is set to an empty
        string, it will read characters as tokens.
        :param s: The input string.
        :type s: str
        """
        if self.delimiter == '': # Character models.
            return list(s.strip())
        else: # Word models.
            return s.strip().split(self.delimiter)

    @timing
    def populate_dictionary(self, filename, vocab, counter, chunk_size):
        with open(filename, encoding=self.encoding) as fin:
            for chunk in tqdm(per_chunk(fin, chunk_size)):
                if all(c == None for c in chunk): break;
                chunk_list_of_tokens = [self.split_tokens(s) for s in chunk if s]
                vocab.add_documents(chunk_list_of_tokens, self.prune_at)
                counter.update(chain(*chunk_list_of_tokens))

    def filter_n_least_frequent(self, vocab, counter, n,
                                keep_tokens=['<pad>', '<s>', '</s>', '<unk>']):
        """
        Remove the least frequent items form the vocabulary.
        :param vocab: self.src_vocab or self.trg_vocab
        :type vocab: gensim.Dictionary
        :param counter: self.src_counter or self.trg_counter
        :type counter: bounter
        :param n: The upper limit of how many items to keep in the vocabulary
        :type n: int
        """
        # If n is bigger than user specified size, don't filter anything.
        if n < len(vocab.token2id):
            good_ids = [vocab.token2id[token] for token, _ in
                       sorted(counter.items(), key=itemgetter(1))[-n:]
                       if token in vocab.token2id]
            good_ids += [self.vocab.token2id[_keep] for _keep in keep_tokens]
            print(good_ids)
            vocab.filter_tokens(good_ids=good_ids)

    def vectorize(self, sent, pad_start=True, pad_end=True):
        """
        Vectorize the sentence, converts it into a list of the indices based on
        the vocabulary. This is used by the `variable_from_sent()`.
        :param sent: The input sentence to convert to vocabulary indices
        :type sent: list(str)
        :param pad_start: Pad the start with the START_IDX [default: True]
        :type pad_end: bool
        :param pad_end: Pad the start with the END_IDX [default: True]
        :type pad_end: bool
        """
        sent = self.split_tokens(sent) if type(sent) == str else sent
        vsent = self.vocab.doc2idx(sent, unknown_word_index=self.UNK_IDX)
        if pad_start:
            vsent = [self.START_IDX] + vsent
        if pad_end:
            vsent = vsent + [self.END_IDX]
        return vsent

    def unvectorize(self, vector, unpad_start=True, unpad_end=True):
        """
        Convert the vector to the natural text sentence.
        """
        return ' '.join([self.vocab[idx] for idx in
                         map(int,chain(*vector))][unpad_left:-unpad_right])

    def reset(self):
        """
        Resets the iterator to the 0th item.
        """
        self.iterable = self._iterate()

    def lines(self):
        """
        The function to iterate through the source and target file.
        """
        with open(self.filename) as fin:
            for line in fin:
                yield line.strip()

    def _iterate(self):
        """
        The helper function to iterate through the source and target file
        and convert the lines into vocabulary indices.
        """
        for line in self.lines():
            sent = self.vectorize(line, self.default_pad_start, self.default_pad_end)
            yield sent

    def __next__(self):
        return next(self.iterable)

    def shuffle(self):
        return iter(sorted(self, key=lambda k: random.random()))
